/* IMPLEMENTATION ARI... VERSION 0.0 OF 26.5.97 BY WH, WU-WIEN & BU-ISTANBUL*/
/* this file:   mainexamp.c */


#define WID 100000 /*samplesize that should be drawn for every distribution*/

#include <math.h>
#include "ari.h"

#define SM_LINT -0x80000000 /*smallest long int on most 32-bit machines*/
#define LA_LINT  0x7FFFFFFF /*largest long int on most 32-bit machines*/

/*structure containing borders and mode of a distribution*/
typedef struct{ long int m/*mode*/,
                 r,l;/*smallest and largest integer with positive porbability*/
      	      }DOMEN;


/*******************************************************************/
/* this is a simple linear congruential uniform generator with
   modulus 2^32 multiplier 69069 and increment 1 as suggested in
   the early 1970ties by G. Marsaglia, the seed is a global variable*/

unsigned long int urn=9917;

double uniform()
{
   urn=(69069*(urn)+1);
   return(urn/4.294967296e9);
}

/******************************************************************/
/*computes log(k!) using the Stirling approximation and a table for
  small values of k*/
double lnkfakt(k)
long int k;
{ static double lnw2p=0.9189385332046727418 ,tabl[10]=
    {0., 0., 0.69314718055994530942, 1.7917594692280550008, 
       3.1780538303479456196,4.7874917427820459942, 6.5792512120101009951, 
       8.5251613610654143, 10.604602902745250228, 12.801827480081469611};
  double dk,edk;
  if (k<=9) return(tabl[k]);
  dk=k; 
  edk=1./dk; 
  return(lnw2p+(dk+0.5)*log(dk)-dk+edk*(1./12.-edk*edk* (1./360.)));
}

/**************************************************************************/
#define P (*param)
/*computes a constant for Poisson histogramm*/
void histpsu(param)
PARAM *param;/*mu= param.d1*/ 
{ P.d2=log(P.d1); }
/*Poisson histogramm: */
/*before using histpsu must be called to compute some constants*/
double histp(param,k)
PARAM *param;/*mu= param.d1*/ 
long int k;
{  return exp(-P.d1+k*P.d2-lnkfakt(k));
}
/*************************************************************************/
/*computes constants for Binomial histogramm*/
void histbinsu(param)
PARAM *param;/*param.i1 n,  param.d1 p*/
{ P.d2=lnkfakt(P.i1);
  P.d3=log(P.d1);
  P.d4=log(1-P.d1);
}
/*histogram of binomial distribution: */
/*before using histbinsu must be called to compute some constants*/
double histbin(param,k)
PARAM *param;/*param.i1 n,  param.d1 p*/
long int k;
{ 
  return exp(P.d2-lnkfakt(k)-lnkfakt(P.i1-k)+
	     k*P.d3+(P.i1-k)*P.d4);
}
/*************************************************************************/
/*computes a constant for Hypergeometric histogramm*/
void histhypsu(param)
PARAM *param;/*param.i1 N,  param.i2 M, param.i3 n*/
{ P.d1=lnkfakt(P.i3)-lnkfakt(P.i1)+lnkfakt(P.i2)+
       lnkfakt(P.i1-P.i2)+lnkfakt(P.i1-P.i3);}
/*Hypergeometr. histogramm:*/
/*before using histhypsu must be called to compute some constants*/
double histhyp(param,k)
PARAM *param;/*param.i1 N,  param.i2 M, param.i3 n*/
long int k;
{  return exp(P.d1-lnkfakt(k)-lnkfakt(P.i2-k)-lnkfakt(P.i3-k)-
	      lnkfakt(P.i1-P.i2-P.i3+k));
}
/*************************************************************************/
/*computes constants for negative Binomial histogramm*/
void histnbsu(param)
PARAM *param;/*param.i1 n,  param.d1 p*/
{ 
  P.d2=P.i1*log(P.d1)-lnkfakt(P.i1-1);
  P.d3=log(1-P.d1);
}
/*Negativ Binomial histogramm:*/
/*before using histnbsu must be called to compute some constants*/
double histnb(param,k)
PARAM *param;/*param.i1 n,  param.d1 p*/
long int k;
{ 
  return exp(lnkfakt(P.i1+k-1)-lnkfakt(k)+P.d2+k*P.d3);
}
/**************************************************************************/
/*histogramm of Zipf-3-distribution, P(k)=const/k^3*/
double histzipf3(param,k)
PARAM *param;
long int k;
{ static double konst=0.831907372581 /* = 1/zeta[3] */;
  return konst*exp(-log((double)k)*3.);
}
/**************************************************************************/
/*proportional to histogramm of Zipf-2-distribution P(k)=const/k^2*/
double histzipf2prop(param,k)
PARAM *param;
long int k;
{ 
  return exp(-log((double)k)*2.);
}
/*************************************************************************/
/*computes for the above histograms the mode and domain of the distribution*/
void compdom(h,domain,param)
DOMEN *domain;
double (*h)();
PARAM *param;/*die Parameter der Verteilung, Aufruf von h als (*h)(param,k)*/
{ long int help;
  (*domain).l=SM_LINT;
  (*domain).r=LA_LINT;
  if (h==histp)     
    { (*domain).m=(long int)(P.d1);
      (*domain).l=0;
    }
  if (h==histbin)
    { (*domain).m=(long int)((P.i1+1)*P.d1);
      (*domain).r=(P.i1);
      (*domain).l=0;
    }
  if (h==histhyp)
    { (*domain).m=(long int)((P.i3+1.)*(P.i2+1.)/(P.i1+2.));
      if (P.i2<=P.i3) (*domain).r=(P.i2);
      else (*domain).r=(P.i3);
      help=P.i3-P.i1+P.i2;
      if (help<=0) (*domain).l=0;
      else (*domain).l=help;
    }
  if (h==histnb)
    { if (P.i1*(1.-P.d1)<=1) (*domain).m=0;
      else (*domain).m=ceil((P.i1*(1.-P.d1)-1.)/P.d1);
      (*domain).l=0;
    }
  if (h==histzipf3||h==histzipf2prop)
    { (*domain).m=1;
      (*domain).l=1;
    }
}
/*************************************************************************/
/*computes the chi-square statistic*/
double chi2discr(classes,obs,wid,domain,h,param,output)
long int classes/*number of classes*/,obs[]/*vector of counted values*/,
     wid/*number of repetitions*/;
DOMEN *domain;
double (*h)();
PARAM *param;/*paramters of the distr., call of the prob(k) as (*h)(param,k)*/
char output/*(0.. no, 1... result, 2.. more)*/;
{ double chi2=0.,ex,sumex=0.,sumob=0.,ob;
  long int i,b,a,cl;
 cl=classes+classes;
 do{ cl=cl/2;
     b=(*domain).m-cl/2;
     if(b<(*domain).l) b=(*domain).l;
     a=b+cl-1;
     if(a>(*domain).r) a=(*domain).r;
   }while((*h)(param,a)*wid<5||(*h)(param,b)*wid<5);


for(i=b;i<a;i++)
  { ex=(*h)(param,i)*wid;
    ob=obs[i];    
    sumex+=ex;
    sumob+=ob;
    chi2+= (ob-ex)*(ob-ex)/ex;
    if(output>=2) printf("for %ld: exp %.2f ob %.0f\n",i,ex,ob); 
  }
  ob=wid-sumob;
  ex=wid-sumex; 
  chi2+= (ob-ex)*(ob-ex)/ex;
  if(output>=2) printf("All other numbers: exp %.2f ob %.0f\n",ex,ob);
  if(output>=1) 
    printf("Chi2-value for samplesize n=%ld  and %ld DF:   %f\n",wid,a-b,chi2);
  return(chi2);
}
/*************************************************************************/



#define ARRSIZE 10000

long kcount[ARRSIZE];

void main()

{ long int i,k;
  PARAM par;
  DOMEN domain;
  void *poi,*poi1;


  prariopt();

  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nPoisson distribution with mu=50.\n");
  par.d1=50.;
  histpsu(&par);
  compdom(histp,&domain,&par);
  poi=arisetup(histp,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=arisample(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;}
  free(poi);
  chi2discr(50,kcount,WID,&domain,histp,&par,1);
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nBinomial distribution with p=0.3 and n=200\n");
  par.d1=0.3;
  par.i1=200;
  histbinsu(&par);
  compdom(histbin,&domain,&par);
  poi=arisetup(histbin,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=arisample(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;
    }
  free(poi);
  chi2discr(50,kcount,WID,&domain,histbin,&par,1);
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nHypergeometric distribution with N=1000 M=100 n=90\n");
  par.i1=1000;
  par.i2=100;
  par.i3=90;
  histhypsu(&par);
  compdom(histhyp,&domain,&par);
  poi=arisetup(histhyp,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=arisample(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;
    }
  free(poi);
  chi2discr(50,kcount,WID,&domain,histhyp,&par,1);
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nNegative binomial distribution with n=3 p=0.7\n");
  par.i1=3;
  par.d1=0.7;
  histnbsu(&par);
  compdom(histnb,&domain,&par);
  poi=arisetup(histnb,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=arisample(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;
    }
  free(poi);
  chi2discr(50,kcount,WID,&domain,histnb,&par,1);
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nZipf 3 distribution\n");
printf("algorithm only correct if transformation with c<=-1/3 is used!\n");
  compdom(histzipf3,&domain,&par);
  poi=arisetup(histzipf3,&par,domain.m,domain.l,domain.r,-1.);
  if(poi!=NULL)/*checks if the set-up was successfull*/
    { for(i=0;i<WID;i++) 
	{  k=arisample(poi);
	   if (k>=0&&k<ARRSIZE) kcount[k]++;}
      free(poi);
      chi2discr(50,kcount,WID,&domain,histzipf3,&par,1);
      for(i=0;i<ARRSIZE;i++) kcount[i]=0;}
printf("\nZipf 2 distribution \n");
printf("algorithm only correct if transformation with c<=-1/2 is used!\n");
/*example for a distribution where only the histogram up to a constant factor
  is implemented. a rough estimation of the modal probability(1/zeta(2))=0.61*/
  compdom(histzipf2prop,&domain,&par);
  poi=arisetup(histzipf2prop,&par,domain.m,domain.l,domain.r,0.61);
  if(poi!=NULL)/*checks if the set-up was successfull*/
    { for(i=0;i<WID;i++) 
	{ k=arisample(poi);
	  if (k>=0&&k<ARRSIZE) kcount[k]++;}
      free(poi);
/* the chi2 test for the Zipf-2 distribution is not possible as the 
   probabilites are computed only up to proportionality.
   Therefore the distribution is compared with the zipf3 distribution and
   the chi2 test-statistic should be large thus showing the bad fit*/
      printf("A high chi2 value is expected as the fit of Zipf(3) is bad\n");
      chi2discr(50,kcount,WID,&domain,histzipf3,&par,1);}

  printf("\n example of generating alternately random variates of the\n");
  printf("Poisson(20) and the Poisson(30) distribution\n");
/*it is no problem that the parameters in par are changed as the parameter
  structure is stored during the set-up together with all constants*/
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
  par.d1=20.;
  histpsu(&par);
  compdom(histp,&domain,&par);
  poi=arisetup(histp,&par,domain.m,domain.l,domain.r,-1.);
  par.d1=30.;
  histpsu(&par);
  compdom(histp,&domain,&par);
  poi1=arisetup(histp,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { if(i%2==0)k=arisample(poi);
      else k=arisample(poi1);
      if (k>=0&&k<ARRSIZE) kcount[k]++;}
  free(poi);
  free(poi1);
printf("A high chi2 value is expected as we compare the match\n");
printf("of this mixed distribution with the Poisson(25) distribution\n");
  par.d1=25.;
  histpsu(&par);
  chi2discr(50,kcount,WID,&domain,histp,&par,1);


}    










